#!/usr/bin/env python
import os
from urllib.parse import urljoin

import requests
from ebi_eva_common_pyutils.config import WritableConfig
from ebi_eva_common_pyutils.logger import AppLogger
from retry import retry

from eva_sub_cli import SUB_CLI_CONFIG_FILE, __version__, SUBMISSION_WS_VAR
from eva_sub_cli.auth import get_auth
from eva_sub_cli.validators.validator import READY_FOR_SUBMISSION_TO_EVA

SUB_CLI_CONFIG_KEY_SUBMISSION_ID = "submission_id"
SUB_CLI_CONFIG_KEY_SUBMISSION_UPLOAD_URL = "submission_upload_url"
SUB_CLI_CONFIG_KEY_COMPLETE = "submission_complete"

SUBMISSION_WS_URL = "https://www.ebi.ac.uk/eva/webservices/submission-ws/submission/"


class StudySubmitter(AppLogger):
    def __init__(self, submission_dir, vcf_files, metadata_file, submission_ws_url=None,
                 submission_config: WritableConfig = None, username=None, password=None):
        self.auth = get_auth(username, password)
        self.submission_ws_url = submission_ws_url
        self.submission_dir = submission_dir
        self.vcf_files = vcf_files
        self.metadata_file = metadata_file
        if submission_config:
            self.sub_config = submission_config
        else:
            config_file = os.path.join(submission_dir, SUB_CLI_CONFIG_FILE)
            self.sub_config = WritableConfig(config_file, version=__version__)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.sub_config.backup()
        self.sub_config.write()

    def _get_submission_ws_url(self):
        """Retrieve the base URL for the submission web services. In order of preference from the user of this class,
        the environment variable or the hardcoded value."""
        if self.submission_ws_url:
            return self.submission_ws_url
        if os.environ.get(SUBMISSION_WS_VAR):
            return os.environ.get(SUBMISSION_WS_VAR)
        else:
            return SUBMISSION_WS_URL

    @property
    def submission_initiate_url(self):
        return self._get_submission_ws_url() + 'submission/initiate'

    @property
    def submission_uploaded_url(self):
        return self._get_submission_ws_url() + 'submission/{submissionId}/uploaded'

    def update_config_with_submission_id_and_upload_url(self, submission_id, upload_url):
        self.sub_config.set(SUB_CLI_CONFIG_KEY_SUBMISSION_ID, value=submission_id)
        self.sub_config.set(SUB_CLI_CONFIG_KEY_SUBMISSION_UPLOAD_URL, value=upload_url)


    def _upload_submission(self):
        if READY_FOR_SUBMISSION_TO_EVA not in self.sub_config or not self.sub_config[READY_FOR_SUBMISSION_TO_EVA]:
            raise Exception(f'There are still validation errors that needs to be addressed. '
                            f'Please review, address and re-validate before uploading.')

        submission_upload_url = self.sub_config[SUB_CLI_CONFIG_KEY_SUBMISSION_UPLOAD_URL]

        for f in self.vcf_files:
            self._upload_file(submission_upload_url, f)
        self._upload_file(submission_upload_url, self.metadata_file)

    @retry(tries=5, delay=10, backoff=5)
    def _upload_file(self, submission_upload_url, input_file):
        base_name = os.path.basename(input_file)
        self.info(f'Transfer {base_name} to EVA FTP')
        r = requests.put(urljoin(submission_upload_url, base_name), data=open(input_file, 'rb'))
        r.raise_for_status()
        self.info(f'Upload of {base_name} completed')

    def verify_submission_dir(self, submission_dir):
        if not os.path.exists(submission_dir):
            os.makedirs(submission_dir)
        if not os.access(submission_dir, os.W_OK):
            raise Exception(f"The directory '{submission_dir}' does not have write permissions.")

    def submit(self, resume=False):
        if READY_FOR_SUBMISSION_TO_EVA not in self.sub_config or not self.sub_config[READY_FOR_SUBMISSION_TO_EVA]:
            raise Exception(f'There are still validation errors that need to be addressed. '
                            f'Please review, address and re-validate before submitting.')
        if not (resume or self.sub_config.get(SUB_CLI_CONFIG_KEY_SUBMISSION_UPLOAD_URL)):
            self.verify_submission_dir(self.submission_dir)
            response = requests.post(self.submission_initiate_url,
                                     headers={'Accept': 'application/hal+json',
                                              'Authorization': 'Bearer ' + self.auth.token})
            response.raise_for_status()
            response_json = response.json()
            self.info(f'Submission ID {response_json["submissionId"]} received!!')
            # update config with submission id and upload url
            self.update_config_with_submission_id_and_upload_url(response_json["submissionId"], response_json["uploadUrl"])

        # upload submission
        self._upload_submission()

        # Complete the submission
        response = requests.put(
            self.submission_uploaded_url.format(submissionId=self.sub_config.get(SUB_CLI_CONFIG_KEY_SUBMISSION_ID)),
            headers={'Accept': 'application/hal+json', 'Authorization': 'Bearer ' + self.auth.token}
        )
        response.raise_for_status()
        self.info("Submission ID {} Complete".format(self.sub_config.get(SUB_CLI_CONFIG_KEY_SUBMISSION_ID)))
        # update config with completion of the submission
        self.sub_config.set(SUB_CLI_CONFIG_KEY_COMPLETE, value=True)

